@@ -12,6 +12,7 @@
 #include <ktar.h>
 #include <kzip.h>
 
+#include <QBuffer>
 #include <QDebug>
 #include <QFileInfo>
 #include <QRegularExpression>
@@ -698,8 +699,11 @@ void KArchiveTest::testTarMaxLength()
 
 void KArchiveTest::testTarGlobalHeader()
 {
-    KTar tar(QFINDTESTDATA(QLatin1String("global_header_test.tar.gz")));
-    QVERIFY2(tar.open(QIODevice::ReadOnly), "global_header_test.tar.gz");
+    const QString fileName = QFINDTESTDATA("data/global_header_test.tar.gz");
+    QVERIFY(!fileName.isEmpty());
+
+    KTar tar(fileName);
+    QVERIFY2(tar.open(QIODevice::ReadOnly), "data/global_header_test.tar.gz");
 
     const KArchiveDirectory *dir = tar.directory();
     QVERIFY(dir != nullptr);
@@ -716,8 +720,11 @@ void KArchiveTest::testTarGlobalHeader()
 
 void KArchiveTest::testTarPrefix()
 {
-    KTar tar(QFINDTESTDATA(QLatin1String("tar_prefix_test.tar.gz")));
-    QVERIFY2(tar.open(QIODevice::ReadOnly), "tar_prefix_test.tar.gz");
+    const QString fileName = QFINDTESTDATA("data/tar_prefix_test.tar.gz");
+    QVERIFY(!fileName.isEmpty());
+
+    KTar tar(fileName);
+    QVERIFY2(tar.open(QIODevice::ReadOnly), "data/tar_prefix_test.tar.gz");
 
     const KArchiveDirectory *dir = tar.directory();
     QVERIFY(dir != nullptr);
@@ -748,8 +755,11 @@ void KArchiveTest::testTarPrefix()
 
 void KArchiveTest::testTarDirectoryForgotten()
 {
-    KTar tar(QFINDTESTDATA(QLatin1String("tar_directory_forgotten.tar.gz")));
-    QVERIFY2(tar.open(QIODevice::ReadOnly), "tar_directory_forgotten.tar.gz");
+    const QString fileName = QFINDTESTDATA("data/tar_directory_forgotten.tar.gz");
+    QVERIFY(!fileName.isEmpty());
+
+    KTar tar(fileName);
+    QVERIFY2(tar.open(QIODevice::ReadOnly), "data/tar_directory_forgotten.tar.gz");
 
     const KArchiveDirectory *dir = tar.directory();
     QVERIFY(dir != nullptr);
@@ -766,7 +776,10 @@ void KArchiveTest::testTarDirectoryForgotten()
 
 void KArchiveTest::testTarEmptyFileMissingDir()
 {
-    KTar tar(QFINDTESTDATA(QLatin1String("tar_emptyfile_missingdir.tar.gz")));
+    const QString fileName = QFINDTESTDATA("data/tar_emptyfile_missingdir.tar.gz");
+    QVERIFY(!fileName.isEmpty());
+
+    KTar tar(fileName);
     QVERIFY(tar.open(QIODevice::ReadOnly));
 
     const KArchiveDirectory *dir = tar.directory();
@@ -784,7 +797,10 @@ void KArchiveTest::testTarEmptyFileMissingDir()
 
 void KArchiveTest::testTarRootDir() // bug 309463
 {
-    KTar tar(QFINDTESTDATA(QLatin1String("tar_rootdir.tar.gz")));
+    const QString fileName = QFINDTESTDATA("data/tar_rootdir.tar.gz");
+    QVERIFY(!fileName.isEmpty());
+
+    KTar tar(fileName);
     QVERIFY2(tar.open(QIODevice::ReadOnly), qPrintable(tar.fileName()));
 
     const KArchiveDirectory *dir = tar.directory();
@@ -839,8 +855,10 @@ void KArchiveTest::testTarShortNonASCIINames() // bug 266141
 #ifdef Q_OS_WIN
     QSKIP("tar test file encoding not windows compatible");
 #endif
+    const QString fileName = QFINDTESTDATA("data/tar_non_ascii_file_name.tar.gz");
+    QVERIFY(!fileName.isEmpty());
 
-    KTar tar(QFINDTESTDATA(QString("tar_non_ascii_file_name.tar.gz")));
+    KTar tar(fileName);
 
     QVERIFY(tar.open(QIODevice::ReadOnly));
     const KArchiveDirectory *dir = tar.directory();
@@ -857,7 +875,10 @@ void KArchiveTest::testTarShortNonASCIINames() // bug 266141
 
 void KArchiveTest::testTarDirectoryTwice() // bug 206994
 {
-    KTar tar(QFINDTESTDATA(QLatin1String("tar_directory_twice.tar.gz")));
+    const QString fileName = QFINDTESTDATA("data/tar_directory_twice.tar.gz");
+    QVERIFY(!fileName.isEmpty());
+
+    KTar tar(fileName);
     QVERIFY(tar.open(QIODevice::ReadOnly));
 
     const KArchiveDirectory *dir = tar.directory();
@@ -880,7 +901,10 @@ void KArchiveTest::testTarIgnoreRelativePathOutsideArchive()
     // outside of the archive directory. For security reasons extractions should only
     // be allowed within the extracted directory as long as not specifically asked.
 
-    KTar tar(QFINDTESTDATA(QLatin1String("tar_relative_path_outside_archive.tar.bz2")));
+    const QString fileName = QFINDTESTDATA("data/tar_relative_path_outside_archive.tar.bz2");
+    QVERIFY(!fileName.isEmpty());
+
+    KTar tar(fileName);
     QVERIFY(tar.open(QIODevice::ReadOnly));
 
     const KArchiveDirectory *dir = tar.directory();
@@ -1226,7 +1250,7 @@ void KArchiveTest::testZipReadRedundantDataDescriptor()
 
 void KArchiveTest::testZipDirectoryPermissions()
 {
-    QString fileName = QFINDTESTDATA("data/dirpermissions.zip");
+    const QString fileName = QFINDTESTDATA("data/dirpermissions.zip");
     QVERIFY(!fileName.isEmpty());
 
     KZip zip(fileName);
@@ -1239,7 +1263,7 @@ void KArchiveTest::testZipDirectoryPermissions()
 
 void KArchiveTest::testZipWithinZip()
 {
-    QString fileName = QFINDTESTDATA("data/zip_within_zip.zip");
+    const QString fileName = QFINDTESTDATA("data/zip_within_zip.zip");
     QVERIFY(!fileName.isEmpty());
 
     KZip zip(fileName);
@@ -1252,9 +1276,34 @@ void KArchiveTest::testZipWithinZip()
     QCOMPARE(listing[1], QString::fromUtf8("mode=100644 path=test.epub type=file size=525154"));
 }
 
+void KArchiveTest::testZipPrependedData()
+{
+    const QString fileName = QFINDTESTDATA("data/zip64_datadescriptor.zip");
+    QVERIFY(!fileName.isEmpty());
+
+    QFile origFile(fileName);
+    QVERIFY2(origFile.open(QIODevice::ReadOnly), qPrintable(origFile.errorString()));
+    auto zipData = origFile.readAll();
+    QVERIFY(zipData.startsWith("PK"));
+
+    zipData.insert(0, QByteArrayLiteral("PrependedPK\x01\x02Garbage"));
+    QBuffer zipBuffer(&zipData);
+
+    KZip zip(&zipBuffer);
+    QVERIFY2(zip.open(QIODevice::ReadOnly), qPrintable(zip.errorString()));
+
+    QCOMPARE(zip.directory()->entries().size(), 1);
+    QCOMPARE(zip.directory()->entries(), QList{QStringLiteral("-")});
+
+    auto entry = zip.directory()->file(QStringLiteral("-"));
+    QVERIFY(entry);
+    QCOMPARE(entry->size(), 4);
+    QCOMPARE(entry->data(), "abcd");
+}
+
 void KArchiveTest::testZipUnusualButValid()
 {
-    QString fileName = QFINDTESTDATA("data/unusual_but_valid_364071.zip");
+    const QString fileName = QFINDTESTDATA("data/unusual_but_valid_364071.zip");
     QVERIFY(!fileName.isEmpty());
 
     KZip zip(fileName);
@@ -1267,7 +1316,7 @@ void KArchiveTest::testZipUnusualButValid()
 
 void KArchiveTest::testZipDuplicateNames()
 {
-    QString fileName = QFINDTESTDATA("data/out.epub");
+    const QString fileName = QFINDTESTDATA("data/out.epub");
     QVERIFY(!fileName.isEmpty());
 
     KZip zip(fileName);
@@ -1286,7 +1335,7 @@ void KArchiveTest::testZipDuplicateNames()
 
 void KArchiveTest::testZip64()
 {
-    QString fileName = QFINDTESTDATA("data/zip64.pkpass");
+    const QString fileName = QFINDTESTDATA("data/zip64.pkpass");
     QVERIFY(!fileName.isEmpty());
 
     KZip zip(fileName);
@@ -1306,9 +1355,167 @@ void KArchiveTest::testZip64()
     QCOMPARE(entry->position(), 8943);
 }
 
+void KArchiveTest::testZipReopenWithoutDoubleDeletion()
+{
+    // testCreateZip must have been run first.
+    KZip zip(s_zipFileName);
+
+    QVERIFY(zip.open(QIODevice::ReadOnly));
+    QVERIFY(zip.close());
+
+    QVERIFY(zip.open(QIODevice::WriteOnly));
+    QVERIFY(zip.close()); // should not crash
+}
+
+void KArchiveTest::testZip64NestedStored()
+{
+    const QString fileName = QFINDTESTDATA("data/dirpermissions.zip");
+    QVERIFY(!fileName.isEmpty());
+    QFile nested(fileName);
+    QVERIFY(nested.open(QIODevice::ReadOnly));
+    auto nestedData = nested.readAll();
+
+    QBuffer archive;
+
+    KZip writer(&archive);
+    QVERIFY(writer.open(QIODevice::WriteOnly));
+    // Write nested zip file
+    writer.setCompression(KZip::NoCompression);
+    QVERIFY(writer.prepareWriting(QStringLiteral("nested.zip"), {}, {}, nestedData.size()));
+    QVERIFY(writer.writeData(nestedData));
+    QVERIFY(writer.finishWriting(nestedData.size()));
+
+    QVERIFY(writer.close());
+
+    KZip reader(&archive);
+    QVERIFY(reader.open(QIODevice::ReadOnly));
+
+    QCOMPARE(reader.directory()->entries().size(), 1);
+    QCOMPARE(reader.directory()->entries(), QList{QStringLiteral("nested.zip")});
+
+    auto entry = reader.directory()->file(QStringLiteral("nested.zip"));
+    QVERIFY(entry);
+    QCOMPARE(entry->size(), 430);
+    QCOMPARE(entry->size(), nestedData.size());
+
+    QVERIFY(reader.close());
+}
+
+void KArchiveTest::testZip64NestedStoredStreamed()
+{
+    const QString fileName = QFINDTESTDATA("data/zip64_nested_stored_streamed.zip");
+    QVERIFY(!fileName.isEmpty());
+
+    KZip zip(fileName);
+    QVERIFY2(zip.open(QIODevice::ReadOnly), qPrintable(zip.errorString()));
+
+    QCOMPARE(zip.directory()->entries().size(), 2);
+
+    auto entry = zip.directory()->file(QStringLiteral("zip64_datadescriptor.zip"));
+    QVERIFY(entry);
+    QCOMPARE(entry->size(), 150);
+
+    entry = zip.directory()->file(QStringLiteral("zip64_end_of_central_directory.zip"));
+    QVERIFY(entry);
+    QCOMPARE(entry->size(), 200);
+
+    QVERIFY(zip.close());
+}
+
+void KArchiveTest::testZip64EndOfCentralDirectory()
+{
+    const QString fileName = QFINDTESTDATA("data/zip64_end_of_central_directory.zip");
+    QVERIFY(!fileName.isEmpty());
+
+    KZip zip(fileName);
+    QVERIFY2(zip.open(QIODevice::ReadOnly), qPrintable(zip.errorString()));
+
+    QCOMPARE(zip.directory()->entries().size(), 1);
+    QCOMPARE(zip.directory()->entries(), QList{QStringLiteral("-")});
+
+    auto entry = zip.directory()->file(QStringLiteral("-"));
+    QVERIFY(entry);
+    QCOMPARE(entry->size(), 4);
+
+    QVERIFY(zip.close());
+}
+
+void KArchiveTest::testZip64DataDescriptor()
+{
+    const QString fileName = QFINDTESTDATA("data/zip64_datadescriptor.zip");
+    QVERIFY(!fileName.isEmpty());
+
+    KZip zip(fileName);
+    QVERIFY2(zip.open(QIODevice::ReadOnly), qPrintable(zip.errorString()));
+
+    QCOMPARE(zip.directory()->entries().size(), 1);
+    QCOMPARE(zip.directory()->entries(), QList{QStringLiteral("-")});
+
+    auto entry = zip.directory()->file(QStringLiteral("-"));
+    QVERIFY(entry);
+    QCOMPARE(entry->size(), 4);
+    QCOMPARE(entry->data(), "abcd");
+
+    QVERIFY(zip.close());
+}
+
+void KArchiveTest::testZip64ExtraZip64Size()
+{
+    const QString fileName = QFINDTESTDATA("data/zip64_extra_zip64_size.zip.gz");
+    QVERIFY(!fileName.isEmpty());
+
+    KCompressionDevice gzfilter(fileName);
+    QVERIFY2(gzfilter.open(QIODevice::ReadOnly), qPrintable(gzfilter.errorString()));
+
+    // Get the actual test data, which is compressed so the repository does
+    // not have to store 4 MByte of binary data
+    auto zipData = gzfilter.read(5 * 1000 * 1000);
+    QVERIFY(zipData.startsWith("PK"));
+
+    QBuffer zipBuffer(&zipData);
+    KZip zip(&zipBuffer);
+    QVERIFY2(zip.open(QIODevice::ReadOnly), qPrintable(zip.errorString()));
+
+    QCOMPARE(zip.directory()->entries().size(), 1);
+    QCOMPARE(zip.directory()->entries(), QList{QStringLiteral("-")});
+
+    auto entry = zip.directory()->file(QStringLiteral("-"));
+    QVERIFY(entry);
+    QCOMPARE(entry->size(), 4404019208);
+
+    auto readDev = std::unique_ptr<QIODevice>(entry->createDevice());
+    auto head = readDev->read(8);
+    QCOMPARE(head, QByteArrayLiteral("abcd\0\0\0\0"));
+    readDev->seek(entry->size() - 8);
+    auto tail = readDev->read(8);
+    QCOMPARE(tail, QByteArrayLiteral("\0\0\0\0ABCD"));
+
+    QVERIFY(zip.close());
+}
+
+void KArchiveTest::testZip64ExtraZip64Offset()
+{
+    const QString fileName = QFINDTESTDATA("data/zip64_extra_zip64_localheader.oxps");
+    QVERIFY(!fileName.isEmpty());
+
+    KZip zip(fileName);
+    QVERIFY2(zip.open(QIODevice::ReadOnly), qPrintable(zip.errorString()));
+
+    QCOMPARE(zip.directory()->entries().size(), 6);
+
+    auto entry = zip.directory()->file(QStringLiteral("_rels/.rels"));
+    QVERIFY(entry);
+    QCOMPARE(entry->size(), 576);
+
+    const auto data = entry->data();
+    QCOMPARE(data.sliced(0, 6), "<?xml ");
+
+    QVERIFY(zip.close());
+}
+
 void KArchiveTest::testRcc()
 {
-    const QString rccFile = QFINDTESTDATA("runtime_resource.rcc"); // was copied from qtbase/tests/auto/corelib/io/qresourceengine
+    const QString rccFile = QFINDTESTDATA("data/runtime_resource.rcc"); // was copied from qtbase/tests/auto/corelib/io/qresourceengine
     QVERIFY(!rccFile.isEmpty());
     KRcc rcc(rccFile);
     QVERIFY(rcc.open(QIODevice::ReadOnly));
@@ -1376,7 +1583,7 @@ void KArchiveTest::cleanupTestCase()
 #if HAVE_XZ_SUPPORT
 
 /**
- * Prepares dataset for archive filter tests
+ * Prepares dataset for 7zip tests
  */
 void KArchiveTest::setup7ZipData()
 {
@@ -1564,6 +1771,103 @@ void KArchiveTest::test7ZipMaxLength()
     // NOTE Cleanup here
     QFile::remove(fileName);
 }
+
+void KArchiveTest::test7ZipNamelessFile()
+{
+    QBuffer archive;
+
+    K7Zip writer(&archive);
+
+    QVERIFY(writer.open(QIODevice::WriteOnly));
+    QVERIFY(writer.writeFile(QString(), "Hello from a nameless file.", 0100644, "azhar", "users"));
+    QVERIFY(writer.close());
+
+    K7Zip reader(&archive);
+
+    QVERIFY(reader.open(QIODevice::ReadOnly));
+
+    const KArchiveDirectory *dir = reader.directory();
+    QVERIFY(dir != nullptr);
+    const QStringList listing = recursiveListEntries(dir, QLatin1String(""), 0);
+
+    QCOMPARE(listing.count(), 1);
+    QCOMPARE(listing[0], "mode=100644 path=contents type=file size=27");
+
+    QVERIFY(reader.close());
+}
+
+void KArchiveTest::test7ZipMultipleNamelessFiles()
+{
+    const QString fileName = QFINDTESTDATA("data/multiple_nameless_files.7z");
+    QVERIFY(!fileName.isEmpty());
+
+    K7Zip k7zip(fileName);
+    QVERIFY2(k7zip.open(QIODevice::ReadOnly), "data/multiple_nameless_files.7z");
+
+    const KArchiveDirectory *dir = k7zip.directory();
+    QVERIFY(dir != nullptr);
+    const QStringList listing = recursiveListEntries(dir, QLatin1String(""), 0);
+
+    QCOMPARE(listing.count(), 2);
+
+    QCOMPARE(listing[0], QString("mode=10600 path=multiple_nameless_files_1 type=file size=27"));
+    QCOMPARE(listing[1], QString("mode=10600 path=multiple_nameless_files_2 type=file size=33"));
+
+    QVERIFY(k7zip.close());
+}
+
+// This test was added to verify K7ZipPrivate::readNumber() boundary check.
+// See: https://invent.kde.org/frameworks/karchive/-/merge_requests/105
+void KArchiveTest::test7ZipReadNumber()
+{
+    const QString fileName = QFINDTESTDATA("data/readnumber_check.7z");
+    QVERIFY(!fileName.isEmpty());
+
+    K7Zip k7zip(fileName);
+    QVERIFY2(k7zip.open(QIODevice::ReadOnly), "data/readnumber_check.7z");
+
+    const KArchiveDirectory *dir = k7zip.directory();
+    QVERIFY(dir != nullptr);
+    const QStringList listing = recursiveListEntries(dir, QLatin1String(""), 0);
+
+    QCOMPARE(listing.count(), 1);
+    QCOMPARE(listing[0], QString("mode=644 path=file.txt type=file size=5"));
+
+    QVERIFY(k7zip.close());
+}
+
+#if HAVE_OPENSSL_SUPPORT
+void KArchiveTest::test7ZipPasswordProtected()
+{
+    const QString fileName = QFINDTESTDATA("data/password_protected.7z");
+    QVERIFY(!fileName.isEmpty());
+
+    K7Zip k7zip(fileName);
+    QVERIFY(!k7zip.open(QIODevice::ReadOnly)); // Should fail without a password
+
+    QVERIFY(k7zip.passwordNeeded());
+    k7zip.setPassword("youshallnotpass");
+
+    QVERIFY2(k7zip.open(QIODevice::ReadOnly), "data/password_protected.7z");
+
+    const KArchiveDirectory *dir = k7zip.directory();
+    QVERIFY(dir != nullptr);
+
+    const KArchiveEntry *entry = dir->entry("file.txt");
+    QVERIFY(entry != nullptr);
+
+    QVERIFY(entry->name() == "file.txt");
+    QVERIFY(entry->isFile());
+
+    const QByteArray fileData = QByteArray("Hello from an encrypted file in a 7z archive!\n");
+
+    const KArchiveFile *file = static_cast<const KArchiveFile *>(entry);
+    QCOMPARE(file->data(), fileData);
+
+    QVERIFY(k7zip.close());
+}
+#endif
+
 #endif
 
 #include "moc_karchivetest.cpp"
