@@ -71,6 +71,18 @@ static uint transformFromMsDos(const char *buffer)
     return dt.toSecsSinceEpoch();
 }
 
+static uint parseUi32(const char *buffer)
+{
+    return uint((uchar)buffer[0] | (uchar)buffer[1] << 8 | (uchar)buffer[2] << 16 | (uchar)buffer[3] << 24);
+}
+
+static quint64 parseUi64(const char *buffer)
+{
+    const uint a = uint((uchar)buffer[0] | (uchar)buffer[1] << 8 | (uchar)buffer[2] << 16 | (uchar)buffer[3] << 24);
+    const uint b = uint((uchar)buffer[4] | (uchar)buffer[5] << 8 | (uchar)buffer[6] << 16 | (uchar)buffer[7] << 24);
+    return (a | (quint64)b << 32);
+}
+
 // == parsing routines for zip headers
 
 /** all relevant information about parsing file information */
@@ -84,7 +96,7 @@ struct ParseFileInfo {
     int uid; // user id (-1 if not specified)
     int gid; // group id (-1 if not specified)
     QByteArray guessed_symlink; // guessed symlink target
-    int extralen; // length of extra field
+    uint dataoffset = 0; // offset from start of local header to data
 
     // parsing related info
     bool exttimestamp_seen; // true if extended timestamp extra field
@@ -95,12 +107,14 @@ struct ParseFileInfo {
     // file sizes from a ZIP64 extra field
     quint64 uncompressedSize = 0;
     quint64 compressedSize = 0;
+    // position of the Local File Header itself, or from the
+    // ZIP64 extra field in the Central Directory
+    quint64 localheaderoffset = 0;
 
     ParseFileInfo()
         : perm(0100644)
         , uid(-1)
         , gid(-1)
-        , extralen(0)
         , exttimestamp_seen(false)
         , newinfounix_seen(false)
     {
@@ -234,18 +248,11 @@ static bool parseInfoZipUnixNew(const char *buffer, int size, bool islocal,
  * parses the extra field
  * @param buffer start of buffer where the extra field is to be found
  * @param size size of the extra field
- * @param islocal true if this is part of a local header, false if of central
  * @param pfi ParseFileInfo object which to write the results into
  * @return true if parsing was successful
  */
-static bool parseExtraField(const char *buffer, int size, bool islocal, ParseFileInfo &pfi)
+static bool parseExtraField(const char *buffer, int size, ParseFileInfo &pfi)
 {
-    // extra field in central directory doesn't contain useful data, so we
-    // don't bother parsing it
-    if (!islocal) {
-        return true;
-    }
-
     while (size >= 4) { // as long as a potential extra field can be read
         int magic = (uchar)buffer[0] | (uchar)buffer[1] << 8;
         buffer += 2;
@@ -262,19 +269,22 @@ static bool parseExtraField(const char *buffer, int size, bool islocal, ParseFil
         switch (magic) {
         case 0x0001: // ZIP64 extended file information
             if (size >= 8) {
-                pfi.uncompressedSize = qFromLittleEndian(*reinterpret_cast<const quint64 *>(buffer));
+                pfi.uncompressedSize = parseUi64(buffer);
             }
             if (size >= 16) {
-                pfi.compressedSize = qFromLittleEndian(*reinterpret_cast<const quint64 *>(buffer + 8));
+                pfi.compressedSize = parseUi64(buffer + 8);
+            }
+            if (size >= 24) {
+                pfi.localheaderoffset = parseUi64(buffer + 16);
             }
             break;
         case 0x5455: // extended timestamp
-            if (!parseExtTimestamp(buffer, fieldsize, islocal, pfi)) {
+            if (!parseExtTimestamp(buffer, fieldsize, true, pfi)) {
                 return false;
             }
             break;
         case 0x5855: // old Info-ZIP unix extra field
-            if (!parseInfoZipUnixOld(buffer, fieldsize, islocal, pfi)) {
+            if (!parseInfoZipUnixOld(buffer, fieldsize, true, pfi)) {
                 return false;
             }
             break;
@@ -297,87 +307,93 @@ static bool parseExtraField(const char *buffer, int size, bool islocal, ParseFil
 }
 
 /**
- * Checks if a token for a central or local header has been found and resets
- * the device to the begin of the token. If a token for the data descriptor is
- * found it is assumed there is a central or local header token starting right
- * behind the data descriptor, and the device is set accordingly to the begin
- * of that token.
- * To be called when a 'P' has been found.
- * @param buffer start of buffer with the 3 bytes behind 'P'
+ * Advance the current file position to the next possible header location
+ *
  * @param dev device that is read from
- * @param dataDescriptor only search for data descriptor
- * @return true if a local or central header begin is or could be reached
+ * @param header storage location for the complete header
+ * @param minSize required size of the header
+ * @return true if a possible header location matching the 'PK' signature
+ *     bytes have been found, false if the end of file has been reached
  */
-static bool handlePossibleHeaderBegin(const char *buffer, QIODevice *dev, bool dataDescriptor)
+static bool seekAnyHeader(QIODevice *dev, QByteArray &header, qsizetype minSize)
 {
-    // we have to detect three magic tokens here:
-    // PK34 for the next local header in case there is no data descriptor
-    // PK12 for the central header in case there is no data descriptor
-    // PK78 for the data descriptor in case it is following the compressed data
-    // TODO: optimize using 32bit const data for comparison instead of byte-wise,
-    // given we run at least on 32bit CPUs
+    header.resize(1024);
+    int n = dev->peek(header.data(), header.size());
 
-    if (buffer[0] == 'K') {
-        if (buffer[1] == 7 && buffer[2] == 8) {
-            // data descriptor token found
-            dev->seek(dev->pos() + 12); // skip the 'data_descriptor'
-            return true;
+    while (n >= minSize) {
+        if (auto i = QByteArrayView(header.data(), n).indexOf("PK"); i >= 0) {
+            dev->seek(dev->pos() + i);
+            if ((i + minSize) < n) {
+                header.remove(0, i);
+                header.resize(minSize);
+                return true;
+            } else {
+                n = dev->peek(header.data(), minSize);
+                header.resize(n);
+                return n >= minSize;
+            }
         }
+        dev->seek(dev->pos() + (n - 1));
+        n = dev->peek(header.data(), header.size());
+    }
+    header.resize(n);
+    return false;
+}
 
-        if (!dataDescriptor
-            && ((buffer[1] == 1 && buffer[2] == 2) //
-                || (buffer[1] == 3 && buffer[2] == 4))) {
-            // central/local header token found
-            dev->seek(dev->pos() - 4);
-            // go back 4 bytes, so that the magic bytes can be found
-            // in the next cycle...
-            return true;
+/**
+ * Advance the current file position to the next header following
+ * a data descriptor
+ *
+ * @param dev device that is read from
+ * @param isZip64 use Zip64 data descriptor layout
+ * @return true if a data descriptor signature has been found, and the
+ *     compressed size matches the current position advance
+ */
+static bool seekPostDataDescriptor(QIODevice *dev, bool isZip64)
+{
+    // Both size fields are uint64 for Zip64, uint32 otherwise
+    const qsizetype descriptorSize = isZip64 ? 24 : 16;
+
+    QByteArray header;
+    auto oldPos = dev->pos();
+
+    while (seekAnyHeader(dev, header, descriptorSize)) {
+        // qCDebug(KArchiveLog) << "Possible data descriptor header at" << dev->pos() << header;
+        if (header.startsWith("PK\x07\x08")) {
+            qint64 compressedSize = isZip64 ? parseUi64(header.constData() + 8) //
+                                            : parseUi32(header.constData() + 8);
+            if (oldPos + compressedSize == dev->pos()) {
+                dev->seek(dev->pos() + descriptorSize);
+                return true;
+            }
+            dev->seek(dev->pos() + 4); // Skip found 'PK\7\8'
+        } else {
+            dev->seek(dev->pos() + 2); // Skip found 'PK'
         }
     }
     return false;
 }
 
 /**
- * Reads the device forwards from the current pos until a token for a central or
- * local header has been found or is to be assumed.
+ * Advance the current file position to the next header
  * @param dev device that is read from
  * @return true if a local or central header token could be reached, false on error
  */
-static bool seekToNextHeaderToken(QIODevice *dev, bool dataDescriptor)
+static bool seekToNextHeaderToken(QIODevice *dev)
 {
-    bool headerTokenFound = false;
-    char buffer[3];
+    QByteArray header;
 
-    while (!headerTokenFound) {
-        int n = dev->read(buffer, 1);
-        if (n < 1) {
-            // qCWarning(KArchiveLog) << "Invalid ZIP file. Unexpected end of file. (#2)";
-            return false;
-        }
-
-        if (buffer[0] != 'P') {
-            continue;
-        }
-
-        n = dev->read(buffer, 3);
-        if (n < 3) {
-            // qCWarning(KArchiveLog) << "Invalid ZIP file. Unexpected end of file. (#3)";
-            return false;
-        }
-
-        if (handlePossibleHeaderBegin(buffer, dev, dataDescriptor)) {
-            headerTokenFound = true;
+    while (seekAnyHeader(dev, header, 4)) {
+        // qCDebug(KArchiveLog) << "Possible header at" << dev->pos() << header;
+        // PK34 for the next local header in case there is no data descriptor
+        // PK12 for the central header in case there is no data descriptor
+        if (header.startsWith("PK\x03\x04") || header.startsWith("PK\x01\x02")) {
+            return true;
         } else {
-            for (int i = 0; i < 3; ++i) {
-                if (buffer[i] == 'P') {
-                    // We have another P character so we must go back a little to check if it is a magic
-                    dev->seek(dev->pos() - 3 + i);
-                    break;
-                }
-            }
+            dev->seek(dev->pos() + 2); // Skip found 'PK'
         }
     }
-    return true;
+    return false;
 }
 
 ////////////////////////////////////////////////////////////////////////
@@ -408,6 +424,8 @@ public:
     // writeonly mode, or it points to the beginning of the central directory.
     // each call to writefile updates this value.
     quint64 m_offset;
+    // Position of the first Local File Header
+    quint64 m_startPos = 0;
 };
 
 KZip::KZip(const QString &fileName)
@@ -445,7 +463,6 @@ bool KZip::openArchive(QIODevice::OpenMode mode)
     // Check that it's a valid ZIP file
     // KArchive::open() opened the underlying device already.
 
-    quint64 offset = 0; // holds offset, where we read
     // contains information gathered from the local file headers
     QHash<QByteArray, ParseFileInfo> pfi_map;
 
@@ -454,6 +471,8 @@ bool KZip::openArchive(QIODevice::OpenMode mode)
     // We set a bool for knowing if we are allowed to skip the start of the file
     bool startOfFile = true;
 
+    qint64 expectedStartPos = 0;
+
     for (;;) { // repeat until 'end of entries' signature is reached
         // qCDebug(KArchiveLog) << "loop starts";
         // qCDebug(KArchiveLog) << "dev->pos() now : " << dev->pos();
@@ -464,33 +483,41 @@ bool KZip::openArchive(QIODevice::OpenMode mode)
             return false;
         }
 
+        if (!memcmp(buffer, "PK\6\6", 4)) { // 'Zip64 end of central directory record'
+            startOfFile = false;
+            break;
+        }
+
         if (!memcmp(buffer, "PK\5\6", 4)) { // 'end of entries'
-            // qCDebug(KArchiveLog) << "PK56 found end of archive";
+            // qCDebug(KArchiveLog) << "PK56 found end of archive at offset" << dev->pos();
             startOfFile = false;
             break;
         }
 
         if (!memcmp(buffer, "PK\3\4", 4)) { // local file header
-            // qCDebug(KArchiveLog) << "PK34 found local file header";
+            // qCDebug(KArchiveLog) << "PK34 found local file header at offset" << dev->pos();
             startOfFile = false;
-            // can this fail ???
-            dev->seek(dev->pos() + 2); // skip 'version needed to extract'
+
+            ParseFileInfo pfi;
+            pfi.localheaderoffset = dev->pos() - 4;
 
             // read static header stuff
-            n = dev->read(buffer, 24);
-            if (n < 24) {
+            n = dev->read(buffer, 26);
+            if (n < 26) {
                 setErrorString(tr("Invalid ZIP file. Unexpected end of file. (Error code: %1)").arg(4));
                 return false;
             }
+            int neededVersion = (uchar)buffer[0] | (uchar)buffer[1] << 8;
+            bool isZip64 = neededVersion >= 45;
 
-            int gpf = (uchar)buffer[0]; // "general purpose flag" not "general protection fault" ;-)
-            int compression_mode = (uchar)buffer[2] | (uchar)buffer[3] << 8;
-            uint mtime = transformFromMsDos(buffer + 4);
+            int gpf = (uchar)buffer[2]; // "general purpose flag" not "general protection fault" ;-)
+            int compression_mode = (uchar)buffer[4] | (uchar)buffer[5] << 8;
+            uint mtime = transformFromMsDos(buffer + 6);
 
-            const qint64 compr_size = uint(uchar(buffer[12])) | uint(uchar(buffer[13])) << 8 | uint(uchar(buffer[14])) << 16 | uint(uchar(buffer[15])) << 24;
-            const qint64 uncomp_size = uint(uchar(buffer[16])) | uint(uchar(buffer[17])) << 8 | uint(uchar(buffer[18])) << 16 | uint(uchar(buffer[19])) << 24;
-            const int namelen = uint(uchar(buffer[20])) | uint(uchar(buffer[21])) << 8;
-            const int extralen = uint(uchar(buffer[22])) | uint(uchar(buffer[23])) << 8;
+            const qint64 compr_size = parseUi32(buffer + 14);
+            const qint64 uncomp_size = parseUi32(buffer + 18);
+            const int namelen = uint(uchar(buffer[22])) | uint(uchar(buffer[23])) << 8;
+            const int extralen = uint(uchar(buffer[24])) | uint(uchar(buffer[25])) << 8;
 
             /*
               qCDebug(KArchiveLog) << "general purpose bit flag: " << gpf;
@@ -512,13 +539,12 @@ bool KZip::openArchive(QIODevice::OpenMode mode)
                 return false;
             }
 
-            ParseFileInfo pfi;
             pfi.mtime = mtime;
+            pfi.dataoffset = 30 + extralen + namelen;
 
             // read and parse the beginning of the extra field,
             // skip rest of extra field in case it is too long
             unsigned int extraFieldEnd = dev->pos() + extralen;
-            pfi.extralen = extralen;
             int handledextralen = qMin(extralen, (int)sizeof buffer);
 
             // if (handledextralen)
@@ -526,7 +552,7 @@ bool KZip::openArchive(QIODevice::OpenMode mode)
 
             n = dev->read(buffer, handledextralen);
             // no error msg necessary as we deliberately truncate the extra field
-            if (!parseExtraField(buffer, n, true, pfi)) {
+            if (!parseExtraField(buffer, n, pfi)) {
                 setErrorString(tr("Invalid ZIP File. Broken ExtraField."));
                 return false;
             }
@@ -540,7 +566,7 @@ bool KZip::openArchive(QIODevice::OpenMode mode)
             if (gpf & 8) {
                 // here we have to read through the compressed data to find
                 // the next PKxx
-                if (!seekToNextHeaderToken(dev, true)) {
+                if (!seekPostDataDescriptor(dev, isZip64)) {
                     setErrorString(tr("Could not seek to next header token"));
                     return false;
                 }
@@ -563,7 +589,7 @@ bool KZip::openArchive(QIODevice::OpenMode mode)
                     if (compr_size > dev->size()) {
                         // here we cannot trust the compressed size, so scan through the compressed
                         // data to find the next header
-                        if (!seekToNextHeaderToken(dev, false)) {
+                        if (!seekToNextHeaderToken(dev)) {
                             setErrorString(tr("Could not seek to next header token"));
                             return false;
                         }
@@ -586,15 +612,19 @@ bool KZip::openArchive(QIODevice::OpenMode mode)
                 if (!foundSignature) {
                     //                     qCDebug(KArchiveLog) << "Testing for optional data descriptor";
                     // read static data descriptor
-                    n = dev->read(buffer, 4);
+                    n = dev->peek(buffer, 4);
                     if (n < 4) {
                         setErrorString(tr("Invalid ZIP file. Unexpected end of file. (#1)"));
                         return false;
                     }
 
-                    if (buffer[0] != 'P' || !handlePossibleHeaderBegin(buffer + 1, dev, false)) {
+                    QByteArrayView currentHead(buffer, 4);
+                    // qCDebug(KArchiveLog) << "Testing for optional data descriptor @ " << dev->pos() << currentHead;
+                    if (currentHead.startsWith("PK\x07\x08")) {
+                        dev->seek(dev->pos() + 16); // skip rest of the 'data_descriptor'
+                    } else if (!(currentHead.startsWith("PK\x03\x04") || currentHead.startsWith("PK\x01\x02"))) {
                         // assume data descriptor without signature
-                        dev->seek(dev->pos() + 8); // skip rest of the 'data_descriptor'
+                        dev->seek(dev->pos() + 12); // skip rest of the 'data_descriptor'
                     }
                 }
 
@@ -606,13 +636,13 @@ bool KZip::openArchive(QIODevice::OpenMode mode)
             }
             pfi_map.insert(fileName, pfi);
         } else if (!memcmp(buffer, "PK\1\2", 4)) { // central block
-            // qCDebug(KArchiveLog) << "PK12 found central block";
+            // qCDebug(KArchiveLog) << "PK12 found central block at offset" << dev->pos();
             startOfFile = false;
 
             // so we reached the central header at the end of the zip file
             // here we get all interesting data out of the central header
             // of a file
-            offset = dev->pos() - 4;
+            auto offset = dev->pos() - 4;
 
             // set offset for appending new files
             if (d->m_offset == 0) {
@@ -625,63 +655,78 @@ bool KZip::openArchive(QIODevice::OpenMode mode)
                 return false;
             }
 
-            // int gpf = (uchar)buffer[9] << 8 | (uchar)buffer[10];
-            // qCDebug(KArchiveLog) << "general purpose flag=" << gpf;
-            // length of the fileName (well, pathname indeed)
-            int namelen = (uchar)buffer[29] << 8 | (uchar)buffer[28];
-            if (namelen <= 0) {
-                setErrorString(tr("Invalid ZIP file, file path name length smaller or equal to zero"));
-                return false;
-            }
-            QByteArray bufferName = dev->read(namelen);
-            if (bufferName.size() < namelen) {
-                // qCWarning(KArchiveLog) << "Invalid ZIP file. Name not completely read";
-            }
-
-            ParseFileInfo pfi = pfi_map.value(bufferName, ParseFileInfo());
-
-            QString name(QFile::decodeName(bufferName));
-
-            // qCDebug(KArchiveLog) << "name: " << name;
-            // only in central header ! see below.
             // length of extra attributes
             int extralen = (uchar)buffer[31] << 8 | (uchar)buffer[30];
             // length of comment for this file
             int commlen = (uchar)buffer[33] << 8 | (uchar)buffer[32];
             // compression method of this file
             int cmethod = (uchar)buffer[11] << 8 | (uchar)buffer[10];
+            // int gpf = (uchar)buffer[9] << 8 | (uchar)buffer[10];
+            // qCDebug(KArchiveLog) << "general purpose flag=" << gpf;
+            // length of the fileName (well, pathname indeed)
+            int namelen = (uchar)buffer[29] << 8 | (uchar)buffer[28];
+            if (namelen <= 0) {
+                setErrorString(tr("Invalid ZIP file, file path name length is zero"));
+                return false;
+            }
+            const int varDataDesiredLength = namelen + extralen;
+            const QByteArray varData = dev->read(varDataDesiredLength);
+            if (varData.length() < varDataDesiredLength) {
+                setErrorString(tr("Invalid ZIP file, unable to read %1 + %2 bytes for filename and extra data at offset %3")
+                                   .arg(namelen, extralen, dev->pos() - varData.size()));
+                return false;
+            }
+
+            ParseFileInfo extrafi;
+            if (extralen) {
+                parseExtraField(varData.constData() + namelen, extralen, extrafi);
+            }
+
+            QByteArray bufferName(varData.constData(), namelen);
+
+            ParseFileInfo pfi = pfi_map.value(bufferName, ParseFileInfo());
+
+            QString name(QFile::decodeName(bufferName));
+
+            // qCDebug(KArchiveLog) << "name: " << name;
 
             // qCDebug(KArchiveLog) << "cmethod: " << cmethod;
             // qCDebug(KArchiveLog) << "extralen: " << extralen;
 
             // crc32 of the file
-            uint crc32 = (uchar)buffer[19] << 24 | (uchar)buffer[18] << 16 | (uchar)buffer[17] << 8 | (uchar)buffer[16];
+            uint crc32 = parseUi32(buffer + 16);
 
             // uncompressed file size
-            quint64 ucsize = uint32_t((uchar)buffer[27] << 24 | (uchar)buffer[26] << 16 | (uchar)buffer[25] << 8 | (uchar)buffer[24]);
+            quint64 ucsize = parseUi32(buffer + 24);
             if (ucsize == 0xFFFFFFFF) {
-                ucsize = pfi.uncompressedSize;
+                ucsize = extrafi.uncompressedSize;
             }
             // compressed file size
-            quint64 csize = uint32_t((uchar)buffer[23] << 24 | (uchar)buffer[22] << 16 | (uchar)buffer[21] << 8 | (uchar)buffer[20]);
+            quint64 csize = parseUi32(buffer + 20);
             if (csize == 0xFFFFFFFF) {
-                csize = pfi.compressedSize;
+                csize = extrafi.compressedSize;
             }
 
             // offset of local header
-            uint localheaderoffset = (uchar)buffer[45] << 24 | (uchar)buffer[44] << 16 | (uchar)buffer[43] << 8 | (uchar)buffer[42];
+            quint64 localheaderoffset = parseUi32(buffer + 42);
+            if (localheaderoffset == 0xFFFFFFFF) {
+                localheaderoffset = extrafi.localheaderoffset;
+            }
 
-            // some clever people use different extra field lengths
-            // in the central header and in the local header... funny.
-            // so we need to get the localextralen to calculate the offset
-            // from localheaderstart to dataoffset
-            int localextralen = pfi.extralen; // FIXME: this will not work if
-            // no local header exists
-
-            // qCDebug(KArchiveLog) << "localextralen: " << localextralen;
+            // qCDebug(KArchiveLog) << "localheader dataoffset: " << pfi.dataoffset;
 
             // offset, where the real data for uncompression starts
-            uint dataoffset = localheaderoffset + 30 + localextralen + namelen; // comment only in central header
+            qint64 dataoffset = localheaderoffset + pfi.dataoffset;
+            if (pfi.localheaderoffset != expectedStartPos + localheaderoffset) {
+                if (pfi.localheaderoffset == d->m_startPos + localheaderoffset) {
+                    qCDebug(KArchiveLog) << "warning:" << d->m_startPos << "extra bytes at beginning of zipfile";
+                    expectedStartPos = d->m_startPos;
+                } else {
+                    setErrorString(tr("Invalid ZIP file, inconsistent Local Header Offset in Central Directory at %1").arg(offset));
+                    return false;
+                }
+                dataoffset = localheaderoffset + pfi.dataoffset + d->m_startPos;
+            }
 
             // qCDebug(KArchiveLog) << "csize: " << csize;
 
@@ -744,14 +789,15 @@ bool KZip::openArchive(QIODevice::OpenMode mode)
 
             if (entry) {
                 if (pos == -1) {
-                    rootDir()->addEntry(entry);
+                    // We don't want to fail opening potentially malformed files, so void the return value
+                    (void)rootDir()->addEntryV2(entry);
                 } else {
                     // In some tar files we can find dir/./file => call cleanPath
                     QString path = QDir::cleanPath(name.left(pos));
                     // Ensure container directory exists, create otherwise
                     KArchiveDirectory *tdir = findOrCreate(path);
                     if (tdir) {
-                        tdir->addEntry(entry);
+                        (void)tdir->addEntryV2(entry);
                     } else {
                         setErrorString(tr("File %1 is in folder %2, but %3 is actually a file.").arg(entryName, path, path));
                         delete entry;
@@ -774,40 +820,27 @@ bool KZip::openArchive(QIODevice::OpenMode mode)
             startOfFile = false;
             bool foundSignature = false;
 
-            while (!foundSignature) {
-                n = dev->read(buffer, 1);
-                if (n < 1) {
-                    setErrorString(tr("Invalid ZIP file. Unexpected end of file."));
-                    return false;
-                }
-
-                if (buffer[0] != 'P') {
-                    continue;
-                }
-
-                n = dev->read(buffer, 3);
-                if (n < 3) {
-                    setErrorString(tr("Invalid ZIP file. Unexpected end of file."));
-                    return false;
-                }
-
+            QByteArray header;
+            while (seekAnyHeader(dev, header, 4)) {
                 // We have to detect the magic token for a local header: PK\003\004
                 /*
                  * Note: we do not need to check the other magics, if the ZIP file has no
                  * local header, then it has not any files!
                  */
-                if (buffer[0] == 'K' && buffer[1] == 3 && buffer[2] == 4) {
+                if (header.startsWith("PK\x03\x04")) {
                     foundSignature = true;
-                    dev->seek(dev->pos() - 4); // go back 4 bytes, so that the magic bytes can be found...
-                } else {
-                    for (int i = 0; i < 3; ++i) {
-                        if (buffer[i] == 'P') {
-                            // We have another P character so we must go back a little to check if it is a magic
-                            dev->seek(dev->pos() - 3 + i);
-                            break;
-                        }
-                    }
+                    d->m_startPos = dev->pos();
+                    break;
                 }
+                if (dev->pos() > 4 * 1024 * 1024) {
+                    break;
+                }
+                dev->seek(dev->pos() + 2); // Skip found 'PK'
+            }
+
+            if (!foundSignature) {
+                setErrorString(tr("Not a ZIP file, no Local File Header found."));
+                return false;
             }
         } else {
             setErrorString(tr("Invalid ZIP file. Unrecognized header at offset %1").arg(dev->pos() - 4));
@@ -835,32 +868,29 @@ bool KZip::closeArchive()
     qint64 centraldiroffset = device()->pos();
     // qCDebug(KArchiveLog) << "closearchive: centraldiroffset: " << centraldiroffset;
     qint64 atbackup = centraldiroffset;
-    QMutableListIterator<KZipFileEntry *> it(d->m_fileList);
-
-    while (it.hasNext()) {
+    for (KZipFileEntry *entry : d->m_fileList) {
         // set crc and compressed size in each local file header
-        it.next();
-        if (!device()->seek(it.value()->headerStart() + 14)) {
+        if (!device()->seek(entry->headerStart() + 14)) {
             setErrorString(tr("Could not seek to next file header: %1").arg(device()->errorString()));
             return false;
         }
         // qCDebug(KArchiveLog) << "closearchive setcrcandcsize: fileName:"
-        //    << it.value()->path()
-        //    << "encoding:" << it.value()->encoding();
+        //    << entry->path()
+        //    << "encoding:" << entry->encoding();
 
-        uLong mycrc = it.value()->crc32();
+        uLong mycrc = entry->crc32();
         buffer[0] = char(mycrc); // crc checksum, at headerStart+14
         buffer[1] = char(mycrc >> 8);
         buffer[2] = char(mycrc >> 16);
         buffer[3] = char(mycrc >> 24);
 
-        int mysize1 = it.value()->compressedSize();
+        int mysize1 = entry->compressedSize();
         buffer[4] = char(mysize1); // compressed file size, at headerStart+18
         buffer[5] = char(mysize1 >> 8);
         buffer[6] = char(mysize1 >> 16);
         buffer[7] = char(mysize1 >> 24);
 
-        int myusize = it.value()->size();
+        int myusize = entry->size();
         buffer[8] = char(myusize); // uncompressed file size, at headerStart+22
         buffer[9] = char(myusize >> 8);
         buffer[10] = char(myusize >> 16);
@@ -873,13 +903,11 @@ bool KZip::closeArchive()
     }
     device()->seek(atbackup);
 
-    it.toFront();
-    while (it.hasNext()) {
-        it.next();
-        // qCDebug(KArchiveLog) << "fileName:" << it.value()->path()
-        //              << "encoding:" << it.value()->encoding();
+    for (KZipFileEntry *entry : d->m_fileList) {
+        // qCDebug(KArchiveLog) << "fileName:" << entry->path()
+        //              << "encoding:" << entry->encoding();
 
-        QByteArray path = QFile::encodeName(it.value()->path());
+        QByteArray path = QFile::encodeName(entry->path());
 
         const int extra_field_len = (d->m_extraField == ModificationTime) ? 9 : 0;
         const int bufferSize = extra_field_len + path.length() + 46;
@@ -899,24 +927,24 @@ bool KZip::closeArchive()
         // memcpy(buffer, head, sizeof(head));
         memmove(buffer, head, sizeof(head));
 
-        buffer[10] = char(it.value()->encoding()); // compression method
-        buffer[11] = char(it.value()->encoding() >> 8);
+        buffer[10] = char(entry->encoding()); // compression method
+        buffer[11] = char(entry->encoding() >> 8);
 
-        transformToMsDos(it.value()->date(), &buffer[12]);
+        transformToMsDos(entry->date(), &buffer[12]);
 
-        uLong mycrc = it.value()->crc32();
+        uLong mycrc = entry->crc32();
         buffer[16] = char(mycrc); // crc checksum
         buffer[17] = char(mycrc >> 8);
         buffer[18] = char(mycrc >> 16);
         buffer[19] = char(mycrc >> 24);
 
-        int mysize1 = it.value()->compressedSize();
+        int mysize1 = entry->compressedSize();
         buffer[20] = char(mysize1); // compressed file size
         buffer[21] = char(mysize1 >> 8);
         buffer[22] = char(mysize1 >> 16);
         buffer[23] = char(mysize1 >> 24);
 
-        int mysize = it.value()->size();
+        int mysize = entry->size();
         buffer[24] = char(mysize); // uncompressed file size
         buffer[25] = char(mysize >> 8);
         buffer[26] = char(mysize >> 16);
@@ -928,10 +956,10 @@ bool KZip::closeArchive()
         buffer[30] = char(extra_field_len);
         buffer[31] = char(extra_field_len >> 8);
 
-        buffer[40] = char(it.value()->permissions());
-        buffer[41] = char(it.value()->permissions() >> 8);
+        buffer[40] = char(entry->permissions());
+        buffer[41] = char(entry->permissions() >> 8);
 
-        int myhst = it.value()->headerStart();
+        int myhst = entry->headerStart();
         buffer[42] = char(myhst); // relative offset of local header
         buffer[43] = char(myhst >> 8);
         buffer[44] = char(myhst >> 16);
@@ -952,7 +980,7 @@ bool KZip::closeArchive()
             extfield[4] = 1 | 2 | 4; // specify flags from local field
             // (unless I misread the spec)
             // provide only modification time
-            unsigned long time = (unsigned long)it.value()->date().toSecsSinceEpoch();
+            unsigned long time = (unsigned long)entry->date().toSecsSinceEpoch();
             extfield[5] = char(time);
             extfield[6] = char(time >> 8);
             extfield[7] = char(time >> 16);
@@ -1087,17 +1115,19 @@ bool KZip::doPrepareWriting(const QString &name,
     // to save, so that we don't have duplicate file entries when viewing the zip
     // with konqi...
     // CAUTION: the old file itself is still in the zip and won't be removed !!!
-    QMutableListIterator<KZipFileEntry *> it(d->m_fileList);
     // qCDebug(KArchiveLog) << "fileName to write: " << name;
-    while (it.hasNext()) {
-        it.next();
-        // qCDebug(KArchiveLog) << "prepfileName: " << it.value()->path();
-        if (name == it.value()->path()) {
+    for (auto it = d->m_fileList.begin(); it != d->m_fileList.end();) {
+        // qCDebug(KArchiveLog) << "prepfileName: " << entry->path();
+        if (name == (*it)->path()) {
             // also remove from the parentDir
-            parentDir->removeEntry(it.value());
-            // qCDebug(KArchiveLog) << "removing following entry: " << it.value()->path();
-            delete it.value();
-            it.remove();
+            if (!parentDir->removeEntryV2(*it)) {
+                return false;
+            }
+            // qCDebug(KArchiveLog) << "removing following entry: " << entry->path();
+            delete *it;
+            it = d->m_fileList.erase(it);
+        } else {
+            it++;
         }
     }
 
